!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_vibration_mod
!> @brief   perform vibrational analysis
!! @authors Kiyoshi Yagi (KY)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_vibration_mod

  use at_dynvars_str_mod
  use at_enefunc_str_mod
  use at_boundary_str_mod
  use at_output_str_mod
  use at_pairlist_str_mod
  use at_vibration_str_mod
  use at_energy_mod
  use at_output_mod
  use at_qmmm_mod
  use molecules_str_mod
  use fileio_mod
  use fileio_control_mod
  use select_mod
  use select_atoms_mod
  use select_atoms_str_mod
  use string_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none
  private

  ! structures
  type, public :: s_vib_info

    ! For vibrational analysis
    integer          :: runmode             = RunModeHarm
    integer          :: nreplica            = 1
    integer          :: qmsave_period       = 1
    character(256)   :: vibatm_select_index = ''
    character(256)   :: output_minfo_atm    = ''
    real(wp)         :: diff_stepsize       = 0.001_wp
    character(256)   :: gridfile            = ''
    character(256)   :: datafile            = ''
    character(256)   :: gridfolderID        = 'a'
    logical          :: dryrun              = .false.
!    real(wp)         :: cutoff              = -1.0

  end type s_vib_info

  ! subroutines
  public  :: show_ctrl_vibration
  public  :: read_ctrl_vibration
  public  :: setup_vibration
  private :: setup_vibatoms
  public  :: run_vib
  private :: harmonic_analysis
  private :: diagonalize
  private :: print_minfo
  private :: print_minfo_grad
  private :: calc_gridpoints

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_vibration
  !> @brief        show usage of VIBRATION section
  !! @authors      KY
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "remd", "rpath"
  !!                          "vib"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_vibration(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode


    return

  end subroutine show_ctrl_vibration
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_vibration
  !> @brief        read VIBRATIOn section in the control file
  !! @authors      KY
  !! @param[in]    handle   : unit number of control file
  !! @param[out]   vib_info : MINIMIZE section in control parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_vibration(handle, vib_info)

    ! parameters
    character(*),            parameter     :: Section = 'Vibration'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_vib_info),        intent(inout) :: vib_info

    ! local
    logical :: found_error


    return

  end subroutine read_ctrl_vibration

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_vibration
  !> @brief        setup vibration information
  !> @authors      KY
  !! @param[in]    vib_info  : VIBRATION section in control parameters
  !! @param[in]    sel_info  : SELECTION section in control parameters
  !! @param[in]    molecule  : molecular information
  !! @param[in]    qmmm      : QM/MM information
  !! @param[out]   vibration : vibration information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_vibration(vib_info, sel_info, molecule, qmmm, vibration)

    ! formal arguments
    type(s_vib_info),        intent(in)    :: vib_info
    type(s_sel_info),        intent(in)    :: sel_info
    type(s_molecule),        intent(in)    :: molecule
    type(s_qmmm),            intent(in)    :: qmmm
    type(s_vibration),       intent(inout) :: vibration


  end subroutine setup_vibration

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_vibatoms
  !> @brief        define atoms subject to vibrational analysis
  !! @authors      KY
  !! @param[in]    vib_info    : vibration input information
  !! @param[in]    sel_info    : selector input information
  !! @param[in]    molecule    : molecular information
  !! @param[out]   vibration    : vibration parameters
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_vibatoms(vib_info, sel_info, molecule, vibration)

    ! formal arguments
    type(s_vib_info),        intent(in) :: vib_info
    type(s_sel_info),        intent(in) :: sel_info
    type(s_molecule),        intent(in) :: molecule
    type(s_vibration),    intent(inout) :: vibration

    ! local variables
    integer                :: igroup, ng_vib, ng_minfo, ng_minfo_bk, natom, &
                              i, j, nn, offset, temp, iatom
    integer, allocatable   :: glist_vib(:), glist_minfo(:), glist_minfo_bk(:)
    type(s_selatoms), allocatable :: selatoms(:)

    integer, parameter :: max_atm_print = 100

    return

  end subroutine setup_vibatoms

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_vib
  !> @brief        perform vibrational analysis
  !! @authors      YA, KY
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions
  !! @param[inout] dynvars     : dynamic variables
  !! @param[inout] vibration   : vibration information
  !! @param[inout] pairlist    : non-bond pair list
  !! @param[inout] boundary    : boundary condition
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_vib(molecule, enefunc, dynvars, vibration, &
                     output, pairlist, boundary)

    ! formal arguments
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_vibration),        intent(inout) :: vibration
    type(s_output),           intent(inout) :: output
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary


    return

  end subroutine run_vib

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    harmonic_analysis
  !> @brief        harmonic vibrational analysis
  !> @authors      KY
  !! @param[inout] molecule  : molecule information
  !! @param[inout] enefunc   : potential energy functions information
  !! @param[inout] dynvars   : dynamic variables information
  !! @param[inout] vibration : vibration information
  !! @param[inout] output    : output information
  !! @param[inout] pairlist  : non-bond pair list information
  !! @param[inout] boundary  : boundary conditions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine harmonic_analysis(molecule, enefunc, dynvars, vibration, &
                          output, pairlist, boundary)

    ! formal arguments
    type(s_molecule),  target, intent(inout) :: molecule
    type(s_enefunc),           intent(inout) :: enefunc
    type(s_dynvars),   target, intent(inout) :: dynvars
    type(s_vibration), target, intent(inout) :: vibration
    type(s_output),            intent(inout) :: output
    type(s_pairlist),          intent(inout) :: pairlist
    type(s_boundary),          intent(inout) :: boundary

    ! local variables
    integer               :: i, j, k, i1, j1, k1, k2, nat, nat3, iatom
    integer               :: count, icount, ncolomn, nc
    real(wp)              :: energy_ref, energy1, energy2, rmsg, delta, x0
    real(wp)              :: dipole_ref(3), dipole1(3), dipole2(3)
    real(wp), allocatable :: grad1(:), grad2(:), sq_mass3(:),   &
                             hessian(:,:), hess_pack(:), hess_mw_pack(:),  &
                             dipole_derivative(:,:), dd_mw(:,:), dd2(:,:), &
                             infrared(:)
    real(wp), pointer     :: coord(:,:), mass(:)
    integer , pointer     :: vibatom_id(:)
    real(wp), allocatable :: grad_ref(:), vec(:,:), omega(:)
    integer               :: replicaid, ierr

    character(256) :: folder, minfo_folder, basename, fname
    character      :: num*5
    character(1)   :: xyz(3)
    logical        :: savefile, ex

    return

  end subroutine harmonic_analysis

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    diagonalize
  !> @brief        diagonalize a symmetric matrix using Lapack (dspevx)
  !! @authors      KY
  !! @param[in]    n : dimension of the matrix
  !! @param[in]    m : number of solution
  !! @param[in]    H(n*(n+1)/2) : The matrix to be diagonalized (upper half)
  !! @param[out]   C(n,m)       : The eigenvectors
  !! @param[out]   E(n)         : The eigenvalues
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine diagonalize(n, m, H, C, E)

    integer :: n,m,m1
    real(wp), dimension(n*(n+1)/2) :: H
    real(wp), dimension(n,m) :: C
    real(wp), dimension(n) :: E

    character :: jobz,range,uplo
    real(wp)  :: vl,vu,abstol
    integer   :: il,iu,ldz,info
    integer,  dimension(:), allocatable :: ifail,iwork
    real(wp), dimension(:), allocatable :: work

  end subroutine diagonalize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    print_minfo
  !> @brief        print the information to minfo format
  !! @authors      KY
  !! @param[in] vibration : vibration information
  !! @param[in] molecule : molecule information
  !! @param[in] dynvars  : dynamic variables information
  !! @param[in] nat      : number of atoms for vibrational analysis
  !! @param[in] energy   : energy
  !! @param[in] grad     : gradient(3*nat)
  !! @param[in] grad     : hessian matrix(3*nat, 3*nat)
  !! @param[in] dipole   : dipole moment (3)
  !! @param[in] dipole_derivative   : dipole moment derivatives (3, nat*3)
  !! @param[in] omega    : harmonic frequency (nat*3)
  !! @param[in] vec      : normal mode displacement vector (nat*3, nat*3)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine print_minfo(vibration, molecule, dynvars, nat, energy, grad, &
                     hess, dipole, dipole_derivative, omega, vec)

    ! formal arguments
    type(s_vibration), target, intent(inout) :: vibration
    type(s_molecule),  target, intent(inout) :: molecule
    type(s_dynvars),   target, intent(inout) :: dynvars

    integer :: nat
    real(wp) :: energy
    real(wp) :: grad(3,nat), hess(nat*3*(nat*3+1)/2)
    real(wp) :: dipole(3), dipole_derivative(3,nat*3)
    real(wp) :: omega(nat*3), vec(nat*3, nat*3)

    ! local variables
    real(wp), pointer     :: coord(:,:)
    integer , pointer     :: vibatom_id(:)

    integer               :: ifile
    integer               :: i, j, k, iatom, ncolumn, isize
    integer               :: atomic_no
    real(wp)              :: mm_mass

    integer               :: nd
    integer, allocatable  :: domain_nat(:), domain_nf(:)
    integer, allocatable  :: domain_idx(:,:)

  end subroutine print_minfo

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    print_minfo_grad
  !> @brief        print the information to minfo format
  !! @authors      KY
  !! @param[in] basename : basename of the minfo file
  !! @param[in] nat      : number of atoms for vibrational analysis
  !! @param[in] energy   : energy
  !! @param[in] grad     : gradient(3*nat)
  !! @param[in] dipole   : dipole moment (3)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine print_minfo_grad(basename, nat, energy, grad, dipole)

    ! formal arguments
    character(256) :: basename
    integer        :: nat
    real(wp)       :: energy, grad(3,nat), dipole(3)

    integer        :: ifile
    integer        :: i, j, k, ncolumn, isize

  end subroutine print_minfo_grad


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_minfo_grad
  !> @brief        read the information to minfo format
  !! @authors      KY
  !! @param[in]    basename : basename of the minfo file
  !! @param[in]    nat      : number of atoms for vibrational analysis
  !! @param[inout] energy   : energy
  !! @param[inout] grad     : gradient(3*nat)
  !! @param[inout] dipole   : dipole moment (3)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_minfo_grad(basename, nat, energy, grad, dipole)

    ! formal arguments
    character(256) :: basename
    integer        :: nat
    real(wp)       :: energy, grad(nat*3), dipole(3)

    integer        :: ifile
    integer        :: i, j, k, ncolumn, isize


  end subroutine read_minfo_grad

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    calc_gridpoints
  !> @brief        calculate the energy and gradient at the grid points
  !> @authors      KY
  !! @param[inout] molecule : molecule information
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] dynvars  : dynamic variables information
  !! @param[inout] vibration : vibration information
  !! @param[inout] pairlist : non-bond pair list information
  !! @param[inout] boundary : boundary conditions information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine calc_gridpoints(molecule, enefunc, dynvars, vibration, &
                     output, pairlist, boundary)

    ! formal arguments
    type(s_molecule),  target, intent(inout) :: molecule
    type(s_enefunc),   target, intent(inout) :: enefunc
    type(s_dynvars),   target, intent(inout) :: dynvars
    type(s_vibration), target, intent(inout) :: vibration
    type(s_output),            intent(inout) :: output
    type(s_pairlist),          intent(inout) :: pairlist
    type(s_boundary),          intent(inout) :: boundary

    ! local
    type(s_qmmm), pointer :: qmmm
    integer               :: nat, nat3
    integer               :: i, ia, j, ierr
    real(wp), pointer     :: coord(:,:)
    integer , pointer     :: vibatom_id(:)

    character(256) :: folder, minfo_folder, datafile, basename, fname, line
    character(4)   :: fid, fnum, label
    character(7)   :: num
    logical        :: savefile
    integer        :: icount, count, replicaid

    integer        :: ifile, idatafile
    logical        :: ex, thisrun_done, restart
    real(wp)       :: energy, grad(3,vibration%vib_natoms), dipole(3)


  end subroutine calc_gridpoints

end module at_vibration_mod

